
import collections
from copy import copy
import itertools


class TransformedDict(collections.MutableMapping):
    """A dictionary that applies an arbitrary key-altering
       function before accessing the keys"""

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key


def FlatDict(ordered=False, case_sensative=True, separator='.'):
    if ordered:
        basetype = collections.OrderedDict
    else:
        basetype = dict

    class _FlatDict(collections.MutableMapping):
        def __init__(self, *args, **kwargs):
            store = basetype(*args, **kwargs)
            self.store = self.__flat__(store)

        def __flat__(self, map_, kpath=''):
            ''' Flattens key in hierarchical dictionary.

            if kpath is provided new map will be returned.
            Otherwise, flat keys will be created within map_

            Arga:
                map_: map to flatten
                kpath: 
            '''
            nmap = basetype()

            prefix = "{}{}".format(kpath, separator) if kpath else ""
            for k, v in list(map_.items()):
                tk = self.__keytransform__(k)
                if isinstance(v, collections.Mapping):
                    fv = self.__flat__(v, kpath="{}{}".format(prefix, tk))
                    nmap.update(fv)
                else:
                    nmap["{}{}".format(prefix, tk)] = v
            return nmap

        def __keytransform__(self, key):
            if not case_sensative:
                if isinstance(key, str):
                    key = key.lower()
                elif isinstance(key, collections.Iterable):
                    key = type(key)(map(lambda x: x.lower(), key))

            if not isinstance(key, str) and isinstance(key, collections.Iterable):
                key = separator.join(key)
            return key

        def __getitem__(self, key):
            key = self.__keytransform__(key)
            return self.store.__getitem__(key)

        def __setitem__(self, key, value):
            key = self.__keytransform__(key)
            if isinstance(value, collections.Mapping):
                value = self.__flat__(value, kpath=key)
                self.store.update(value)
                return
            self.store.__setitem__(key, value)

        def __delitem__(self, key):
            del self.store[self.__keytransform__(key)]

        def __iter__(self):
            return iter(self.store)

        def __len__(self):
            return len(self.store)

        def __repr__(self):
            return repr(self.store)

        def __str__(self):
            return str(self.store)

        def __copy__(self):
            return type(self)(self.store)

        def copy(self):
            return self.__copy__()

        def update(self, iter_):
            if isinstance(iter_, collections.Mapping):
                iter_ = iter_.items()
            for k, v in iter_:
                self.__setitem__(k, v)

        def asdict(self):
            return self.store

        def _unflat(self, store=None):

            if not store:
                store = self.store

            keyitems = []
            valitems = []
            for k, v in store.items():
                kpath, _, key = k.partition(separator)
                if key:
                    keyitems.append((kpath, (key, v)))
                else:
                    valitems.append((kpath, v))

            keyfunc = lambda x: x[0]
            data = sorted(keyitems, key=keyfunc)
            store = basetype()
            for k, g in itertools.groupby(data, keyfunc):
                subdict = dict([v for _, v in g])
                store[k] = self._unflat(subdict)

            store.update(dict(valitems))
            return store

        def unflat(self):
            ''' Returns self as hierarchical dict of self.store
            '''
            return self._unflat()

    return _FlatDict


def ComplexDict(ordered=False, case_sensative=True):
    if ordered:
        base = collections.OrderedDict
    else:
        base = dict

    class _ComplexDict(base):
        def __keytransform__(self, key):
            if not case_sensative:
                if isinstance(key, str):
                    key = key.lower()
                elif isinstance(key, collections.Iterable):
                    key = type(key)(map(lambda x: x.lower(), key))
            return key

        def __getitem__(self, key):
            if not case_sensative:
                key = key.lower()
            return self[key]

        def __setitem__(self, key, value):
            if not case_sensative:
                key = key.lower()
            self[key] = value

        def flat(self):
            return 

    return _ComplexDict


if __name__ == '__main__':
    import pprint
    from itertools import combinations

    d = { 'LONLYITEM': 'JUSTME',
         "EVENTOR": {
            "DATABASES": {
                "default": {
                    "dialect": "sqlite",
                    "query": {
                        "cache": "shared"
                    },
                },
                "sqfile00": {
                    "dialect": "sqlite",
                    "database": "/var/acrisel/sand/eventor/eventor/eventor/examples/example00.db"  
                },
                "pgdb1": {
                    "dialect": "postgresql",
                    "drivername" : "psycopg2",
                    "username": "arnon",
                    "password": "arnon42",
                    "host": "localhost",
                    "port": 5433,
                    "database": "pyground",
                    "schema": "play",
                },
                "pgdb2": {
                    "dialect": "postgresql",
                    "drivername" : "psycopg2",
                    "username": "arnon",
                    "password": "Chompi42",
                    "host": "192.168.1.70",
                    "port": 5432,
                    "database": "pyground",
                    "schema": "play",
                },
            }}}

    fd = FlatDict()(d)
    print(repr(fd))

    key = 'EVENTOR.DATABASEs.pgdb1.drivername'
    for ordered, case_sensative in set(combinations([True, False, True, False], 2)):
        print("ordered, case_sensative:", ordered, case_sensative)
        f = FlatDict(ordered=ordered, case_sensative=case_sensative, separator='.')(d)
        try:
            print(f[key])
        except KeyError:
            print('Missing:', key)

        f['A'] = {'1': 2, '2': 3}
        f.update({'B': {'1': 2, '2': 3}})
        pprint.pprint(f['B.1'])

    print('LONLYITEM:', f['LONLYITEM'])
    pprint.pprint(f.unflat())    


# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Jun 29, 2016

@author: arnon
'''
import pprint
import json
from copy import deepcopy

class Node(object):
    def __init__(self, name, value=None):
        self._name = name
        self._value = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v

    @value.getter
    def value(self):
        return self._value
        
    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return repr(self.name) + ' : ' + repr(self.value)

    def __str__(self):
        return repr(self)


class Edge(object):
    def __init__(self, node1, node2):
        self.source = node1
        self.target = node2

    def from_node(self):
        return self.source

    def to_node(self):
        return self.target


class DiGraph(object):
    def __init__(self, nodes=None):
        self.nodes = nodes if nodes else dict()
        # self.dfs_weights=dict()

    Node = Node

    def add_node(self, node):
        found = self.nodes.get(node.name)
        if not found:
            result = dict([('node', node),
                           ('nexts', list()),
                           ('prevs', list())])
            self.nodes[node.name] = result
        return found is None
    
    def node_value(self, name):
        #print('Type:',type(self.nodes[name]['node']))
        return self.nodes[name]['node'].value

    
    def prevs(self,node):
        props = self.nodes.get(node.name)
        return props['prevs']
    
    def prevs_by_name(self,node_name):
        props = self.nodes.get(node_name)
        return self.prevs(props['node'])
        
    def add_edge(self, from_node, *to_nodes):
        fname = from_node.name
        fnode = self.nodes.get(fname)
        if not fnode:
            raise Exception("EdgeError: From-Node {} not in tree".format(from_node.name))
            #fnode=self.Node(from_node)
            #self.add_node(from_node)

        nexts = fnode['nexts']
        for node in to_nodes:
            tname = node.name
            tnode = self.nodes.get(node.name)
            if not tnode: 
                raise Exception("EdgeError: To-Node {} not in tree".format(node.name))

            if tname not in nexts:
                nexts.append(tname)
            prevs = tnode['prevs']
            if fname not in prevs:
                prevs.append(fname)
            #print('Added Edge:', from_node, node)
    
    def add_edge_by_name(self,from_name,*to_name):
        from_node=self.nodes.get(from_name)
        to_node=[self.nodes[name]['node'] for name in to_name]
        self.add_edge(from_node['node'],*to_node)
            
        
    def del_edge(self, from_node, *to_nodes):
        fname = from_node.name
        fnode = self.nodes.get(fname)
        nexts = fnode['nexts']
        for node in to_nodes:
            tname = node.name
            try: 
                nexts.remove(tname)
            except Exception: 
                pass

            tnode = self.nodes.get(tname)
            prevs = tnode['prevs']
            try:
                prevs.remove(fname)
            except:
                pass

    def del_node(self, node):
        dname = node.name
        dnode = self.nodes.get(dname)
        nexts = dnode['nexts']
        for tname in nexts:
            tnode = self.nodes.get(tname)
            prevs = tnode['prevs']
            try:
                prevs.remove(dname)
            except Exception:
                pass

        prevs = dnode['prevs']
        for fname in prevs:
            fnode = self.nodes.get(fname)
            nexts = fnode['nexts']
            try:
                nexts.remove(dname)
            except Exception:
                pass
        del self.nodes[dname]

    def _dfs(self, node_name, weights):
        props = self.nodes[node_name]
        try:
            for name in props['nexts']:
                if name not in weights.keys():
                    self._dfs(name, weights)
            weight = 1 + max([weights[name] for name in props['nexts']])
        except Exception:
            weight = 1
        weights[node_name] = weight

    def __len__(self):
        return len(self.nodes)

    def dfs(self, reverse=False):
        ''' Node iterator sorted depth first '''
        weights = dict()
        for node_name, props in self.nodes.items():
            if node_name not in weights.keys():
                self._dfs(node_name, weights)
        for name in sorted(weights.keys(), key=lambda x: weights[x], reverse = not reverse):
            yield self.nodes[name]['node']
    
    def roots(self):
        ''' Find independent nodes
            Return list of nodes with no previous nodes '''
        ans = [name for name,node in self.nodes.items() if not node['prevs']]
        return ans      
        
    def _ddfs(self, weights,graph_clone,roots,max_weight=0):
        for root in roots:
            weights[root] = max_weight
            ''' Get node object from dictionary of nodes and remove it from graph'''
            props = graph_clone.nodes[root]
            graph_clone.del_node(props['node'])
        
        roots = graph_clone.roots()
        if roots:
            self._ddfs(weights, graph_clone, roots, max_weight-1)


    def ddfs(self, reverse=False):
        ''' Node iterator sorted dependent depth first search
            The function returns nodes in the order of dependency. 
            First root nodes, then nodes with no incoming edges after removal of roots, ..... 
        '''
        weights = dict()
        graph_clone = deepcopy(self)
        roots = self.roots()
        self._ddfs(weights,graph_clone,roots)
        
        for name in sorted(weights.keys(), key=lambda x: weights[x], reverse = not reverse):
            yield self.nodes[name]['node']

    def __str__(self):
        return pprint.pformat(self.nodes)

    def to_file(self, fp):
        json.dump(self.nodes, fp)

    @classmethod
    def from_file(cls, fp):
        try:
            nodes = json.load(fp)
        except Exception:
            return None
        else:
            return cls(nodes)


if __name__ == '__main__':
    g = DiGraph()

    a = g.Node(('a', 1), 35)
    b = g.Node(('b', 2), 46)
    c = g.Node(('c', 3), 78)
    d = g.Node(('d', 4), 89)
    e = g.Node(('e', 1), 35)
    
    g.add_node(a)
    g.add_node(b)
    g.add_node(c)
    g.add_node(d)
    g.add_node(e)
    
    g.add_edge_by_name(('a',1), ('b',2))
    g.add_edge(b, c)
    g.add_edge(e, d)
    g.add_edge(c, d)

    print('Resulting graph:', g)

    #g.del_node(d)
    #print(g)
    print('Graph nodes:')
    for node in g.dfs(reverse=True):
        print(node.name, node.value)
    print('Graph roots:', g.roots())
    print('DDFS Nodes iterator:',list(g.ddfs()))
    #print('Original graph:',g)
    print('Edges to d:',g.prevs_by_name(('d',4)))
    print("Retrieve Node's %s value by node name. Value:%s" % (('d',4),g.node_value(('d', 4))))
    


#!/usr/bin/env python3
'''
Created on Jan 2, 2018

@author: arnon
'''

from subprocess import run, PIPE


def install_required(exact=False):
    cmd = ['pip', 'freeze']
    result = run(cmd, shell=False, stdout=PIPE, stderr=PIPE)

    if result.returncode != 0:
        raise RuntimeError("Failed pip freeze: {}".format(result.stderr))

    relations = ">=" if not exact else "=="
    items = []
    for item in result.stdout.decode().split('\n'):
        item = item.strip()
        if item:
            package, _, version = item.partition('==')
            items += ["{}{}{}".format(package, relations, version)]

    return items


def format_setup_required(required, ident=''):
    items = []
    prefix = '{}install_required = ['.format(ident)
    plen = len(prefix) + len(ident)
    for item in required:
        items += ["{}'{}',".format(prefix, item)]
        prefix = ' ' * plen
    items += [']']
    return '\n'.join(items)


def format_file_required(required, file='REQUIRED'):
    with open(file, 'w') as f:
        for require in required:
            f.write("{}\n".format(require))





# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
'''
Created on Sep 21, 2017

@author: arnon
'''
import logging
from copy import copy
from datetime import datetime
from acrilib import get_hostname, get_ip_address
import logging


def logging_record_add_host():
    old_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        if not hasattr(record, 'host'):
            record.host = get_hostname()
        if not hasattr(record, 'ip'):
            record.ip = get_ip_address()
        return record

    logging.setLogRecordFactory(record_factory)


class LoggerAddHostFilter(logging.Filter):
    """
    This is filter adds host information to LogRecord.
    """

    def filter(self, record):

        if not hasattr(record, 'host'):
            record.host = get_hostname()
        if not hasattr(record, 'ip'):
            record.ip = get_ip_address()
        return True


class MicrosecondsDatetimeFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        ct = datetime.fromtimestamp(record.created)

        if ct is None:
            ct = datetime.now()

        if datefmt is not None:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s.%03d" % (t, record.msecs)

        return s


class LevelBasedFormatter(logging.Formatter):
    '''
    defaults={
        logging.DEBUG : u"%(asctime)-15s: %(process)-7s: %(levelname)-7s: %(message)s: %(module)s.%(funcName)s(%(lineno)d)",
        'default' : u"%(asctime)-15s: %(process)-7s: %(levelname)-7s: %(message)s",
        }
    '''

    defaults = {
        logging.DEBUG: (u'[ %(asctime)-15s ][ %(levelname)-7s ][ %(host)s ]'
                        '[ %(processName)-11s ][ %(message)s ]'
                        '[ %(module)s.%(funcName)s(%(lineno)d) ]'),
        'default': (u'[ %(asctime)-15s ][ %(levelname)-7s ][ %(host)s ]'
                  '[ %(processName)-11s ][ %(message)s ]'),
        }

    def __init__(self, level_formats={}, datefmt=None):
        super(LevelBasedFormatter, self).__init__()
        formats = LevelBasedFormatter.defaults
        if level_formats:
            formats = copy(LevelBasedFormatter.defaults)
            formats.update(level_formats)

        self.datefmt = datefmt
        self.formats = dict([(level, MicrosecondsDatetimeFormatter(fmt=fmt, datefmt=self.datefmt))
                             for level, fmt in formats.items()])
        self.default_format = self.formats['default']

    def format(self, record):
        formatter = self.formats.get(record.levelno, self.default_format,)
        try:
            result = formatter.format(record)
        except Exception:
            raise
        return result



#!/usr/bin/env python3
'''
Created on Jan 2, 2018

@author: arnon
'''

from subprocess import run, PIPE
from acrilib.setup.install_required import install_required, format_file_required, format_setup_required


def cmdargs():
    import argparse

    parser = argparse.ArgumentParser(
        description='''produce required packages to add into setup.py.

Example:
    install_required
    ''')

    parser.add_argument('--exact', action='store_true', dest='exact', default=False,
                        help='set relationships as ==, otherwise >= will be used.')
    parser.add_argument('--embed', action='store_true', dest='embed', default=False,
                        help='format for embedding directly into setup.py.')
    parser.add_argument('--output', type=str, dest='output', required=False, default='REQUIRED',
                        help='override the default "REQUIRED" file.')
    args = parser.parse_args()
    return args


def main(args):
    required = install_required(args.exact)
    if not args.embed:
        msg = format_file_required(required, file=args.output)
    else:
        msg = format_setup_required(required)
        print(msg)


if __name__ == '__main__':
    args = cmdargs()
    main(args)

    